###############################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
# OLI Systems, Inc. Copyright © 2022, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of OLI Systems, Inc. nor the names of any contributors to
# the software made available herein may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# You are under no obligation whatsoever to provide any bug fixes, patches, or upgrades to the
# features, functionality or performance of the source code ("Enhancements") to anyone; however,
# if you choose to make your Enhancements available either publicly, or directly to OLI Systems, Inc.,
# without imposing a separate written license agreement for such Enhancements, then you hereby grant
# the following license: a non-exclusive, royalty-free perpetual license to install, use, modify, prepare
# derivative works, incorporate into other computer software, distribute, and sublicense such enhancements
# or derivative works thereof, in binary and source code form.
###############################################################################

__author__ = "Adam Atia, Adi Bannady, Paul Vecchiarelli"


from os.path import isfile, islink
import requests  # from requests import get, post, request
import json
import time

from watertap.tools.oli_api.util.watertap_to_oli_helper_functions import get_oli_name


class OLIApi:
    """
    A class to wrap OLI Cloud API calls and access functions for interfacing with WaterTAP.
    """

    def __init__(self, credential_manager, test=False):
        """
        Constructs all necessary attributes for OLIApi class

        :param credential_manager_class: class used to manage credentials
        :param test: bool switch for automation during tests
        """

        self.test = test
        self.credential_manager = credential_manager
        self.request_auto_login = self.credential_manager.request_auto_login

    # binds OLIApi instance to context manager
    def __enter__(self):
        self.session_dbs_files = []
        return self

    # return False if no exceptions raised
    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        # delete all .dbs files created during session
        for file in self.session_dbs_files:
            self._delete_dbs_file(file)
        return False

    def get_dbs_file_id(self, chemistry_source=None, phases=None, model_name=None):
        """
        Gets dbs_file_id for a given input.

        :param chemistry_source: Path (str), dict, or state block containing chemistry info
        :param phases: Container (dict) for chemistry model parameters
        :param model_name: Name (str) of model OLI will use

        :return dbs_file_id: string name for DBS file ID
        """

        if isinstance(chemistry_source, str):
            if not isfile(chemistry_source) and not islink(chemistry_source):
                raise OSError(
                    "Could not find requested path to file. Please "
                    "check that this path to file exists."
                )
            dbs_file_id = self._upload_dbs_file(chemistry_source)
        else:
            dbs_dict = self._create_dbs_dict(chemistry_source, phases, model_name)
            if not bool(dbs_dict):
                raise RuntimeError(
                    "DBS file generation failed. Ensure your inputs, e.g.,"
                    + "solute names are formatted correctly."
                )
            else:
                dbs_file_id = self._generate_chemistry_file(dbs_dict)

        return dbs_file_id

    def _upload_dbs_file(self, dbs_file_path):
        """
        Uploads a dbs file to the OLI Cloud given a full file path.

        :param file_path: full path to dbs file

        :return dbs_file_id: string name for DBS file ID
        """

        try:
            with open(dbs_file_path, "rb") as file:
                files = {"files": file}
                req_result = self.request_auto_login(
                    lambda headers: requests.post(
                        self.credential_manager.upload_dbs_url,
                        headers=headers,
                        files=files,
                    )
                )
            dbs_file_id = req_result["file"][0]["id"]
            self.session_dbs_files.append(dbs_file_id)
            return dbs_file_id
        except:
            raise IOError(" Failed to upload DBS file. Ensure specified file exists.")

    def _create_dbs_dict(self, chemistry_source, phases, model_name):

        """
        Creates dict for chemistry-builder to later generate a DBS file id.

        :param chemistry_source: OLI-compatible ion names as a Pyomo Set, list, or dict,
            where the keys are ion names
        :param model_name: Name for model as a string
        :param phases: OLI-compatible phases

        :return dbs_dict: dict containing params for DBS file generation
        """

        # TODO: enable direct use of state block (via helper functions)
        if not isinstance(chemistry_source, (list, dict)):
            raise IOError(
                " Provide a list, dict, or Pyomo set of OLI-compatible solute names (e.g., NAION"
            )
        if len(chemistry_source) == 0:
            raise IOError(
                f" Unable to create DBS dict from empty {type(chemistry_source)}."
            )
        else:
            try:
                solute_list = [
                    {"name": get_oli_name(solute)} for solute in chemistry_source
                ]
            except AttributeError:
                solute_list = [{"name": solute.oli_name} for solute in chemistry_source]

        if model_name is None:
            model_name = input("Enter model name: ")
        if phases is None:
            phases = ["liquid1", "solid"]
        params = {
            "thermodynamicFramework": "MSE (H3O+ ion)",
            "modelName": model_name,
            "phases": phases,
            "inflows": solute_list,
        }
        dbs_dict = {"method": "chemistrybuilder.generateDBS", "params": params}
        return dbs_dict

    def _generate_chemistry_file(self, dbs_dict):
        """
        :param dbs_dict: dict containing params for DBS file generation

        :return dbs_file_id: string name for DBS file ID
        """

        if (dbs_dict is None) or (len(dbs_dict) == 0):
            raise IOError(
                f" Invalid container {dbs_dict} for chemistry file generation call."
            )
        else:

            def add_additional_header(headers):
                headers["content-type"] = "application/json"
                return requests.post(
                    url=self.credential_manager.dbs_url,
                    headers=headers,
                    data=json.dumps(dbs_dict),
                )

            req_result = self.request_auto_login(add_additional_header)
            if req_result["status"] == "SUCCESS":
                dbs_file_id = req_result["data"]["id"]
                self.session_dbs_files.append(dbs_file_id)
                return dbs_file_id
            raise RuntimeError(" Failed to generate chemistry file.")

    def get_user_summary(self):
        """
        Gets information for all files on user's cloud.

        :return user_summary: dictionary containing file information and flash history for each dbs file
        """

        user_summary = {}
        for entry in self.get_user_dbs_files()["data"]:
            dbs_file_id = entry["fileId"]
            user_summary[dbs_file_id] = {
                "dbs_file_info": self.get_dbs_file_info(dbs_file_id=dbs_file_id),
                "flash_history": self.get_flash_history(dbs_file_id=dbs_file_id),
                "job_id": self.get_flash_history(dbs_file_id=dbs_file_id, job_id=True),
            }
        return user_summary

    def get_user_dbs_files(self):
        """
        Gets all DBS files on user's cloud.

        :return user_dbs_files: dictionary containing a list of all user DBS files uploaded
        """

        user_dbs_files = self.request_auto_login(
            lambda headers: requests.get(
                self.credential_manager.dbs_url, headers=headers
            )
        )
        return user_dbs_files

    def get_dbs_file_info(self, dbs_file_id=""):
        """
        Retrieves chemistry information from OLI Cloud.

        :param dbs_file_id: string ID of DBS file

        :return dbs_file_info: dictionary containing information about the DBS file
        """

        endpoint = (
            f"{self.credential_manager.engine_url}file/{dbs_file_id}/chemistry-info"
        )

        def add_additional_header(headers):
            headers["content-type"] = "application/json"
            r = requests.get(endpoint, headers=headers)
            return r

        dbs_file_info = self.request_auto_login(add_additional_header)
        return dbs_file_info

    def get_flash_history(self, dbs_file_id, job_id=False):
        """
        Retrieves history of flash information, e.g., input for a chemistry model.

        :param dbs_file_id: string ID of DBS file
        :param job_id: bool flag to return job ID for troubleshooting

        :return flash_history: dictionary containing array of submitted jobs OR just job ID if specified
        """

        endpoint = f"{self.credential_manager.engine_url}/flash/history/{dbs_file_id}"
        flash_history = self.request_auto_login(
            lambda headers: requests.get(endpoint, headers=headers)
        )
        if job_id:
            return flash_history["data"][0]["jobId"]
        return flash_history

    def dbs_file_cleanup(self, dbs_file_ids=None):
        """
        Deletes all (or specified) DBS files on OLI Cloud.

        :param dbs_file_ids: list of DBS files that should be deleted
        """

        if dbs_file_ids is None:
            dbs_file_ids = [
                entry["fileId"] for entry in self.get_user_dbs_files()["data"]
            ]
        if self._delete_permission(dbs_file_ids):
            for dbs_file_id in dbs_file_ids:
                self._delete_dbs_file(dbs_file_id)

    def _delete_permission(self, dbs_file_ids):
        """
        Ensures user permits deletion of specified files.

        :param dbs_file_ids: list of DBS files that should be deleted

        :return boolean: status of user permission (to delete DBS files from OLI Cloud)
        """

        if self.test is True:
            return True
        else:
            print("WaterTAP will delete dbs files:")
            for file in dbs_file_ids:
                print(f"- {file}")
            print("from OLI Cloud.")
            r = input("[y]/n: ")
            if (r.lower() == "y") or (r == ""):
                return True
            return False

    def _delete_dbs_file(self, dbs_file_id):
        """
        Deletes a DBS file.

        :param dbs_file_id: string ID of DBS file
        """

        endpoint = f"{self.credential_manager._delete_dbs_url}{dbs_file_id}"
        if self.credential_manager.access_key:
            headers = {"authorization": "API-KEY " + self.credential_manager.access_key}
        else:
            headers = {"authorization": "Bearer " + self.credential_manager.jwt_token}
        requests.request("DELETE", endpoint, headers=headers, data={})

    def call(
        self,
        function_name,
        dbs_file_id,
        json_input=dict(),
        poll_time=1.0,
        max_request=1000,
        tee=False,
    ):
        """
        Calls a function in the OLI Engine API.

        :param function_name: name of function to call
        :param dbs_file_id: string ID of DBS file
        :param json_input: calculation input JSON
        :param poll_time: max delay between each call
        :param max_request: maximum requests
        :param tee: boolean argument to hide or display print messages

        :return: dictionary containing result or error
        """

        endpoint = ""
        method = "POST"
        if function_name == "corrosion-contact-surface":
            endpoint = f"{self.credential_manager.engine_url}file/{dbs_file_id}/{function_name}"
            method = "GET"
        else:
            endpoint = f"{self.credential_manager.engine_url}flash/{dbs_file_id}/{function_name}"
            method = "POST"

        if bool(json_input):
            data = json.dumps(json_input)
        else:
            # TODO: raise error
            data = ""

        def add_additional_header(headers):
            headers["content-type"] = "application/json"
            if method == "POST":
                return requests.post(endpoint, headers=headers, data=data)
            output = requests.get(endpoint, headers=headers, data=data)
            return output

        results_link = ""
        start_time = time.time()
        request_result1 = self.request_auto_login(add_additional_header)
        end_time = time.time()
        request_time = end_time - start_time
        if tee:
            print("First request time =", request_time)
        if bool(request_result1):
            if request_result1["status"] == "SUCCESS":
                if "data" in request_result1:
                    if "status" in request_result1["data"]:
                        if (
                            request_result1["data"]["status"] == "IN QUEUE"
                            or request_result1["data"]["status"] == "IN PROGRESS"
                        ):
                            if "resultsLink" in request_result1["data"]:
                                results_link = request_result1["data"]["resultsLink"]
        if tee:
            print(results_link)
        # TODO: raise error
        if results_link == "":
            return dict()

        # poll on results link until success
        data = ""
        endpoint = results_link
        method = "GET"
        request_iter = 0

        while True:
            # make request and time
            start_time = time.time()
            request_result2 = self.request_auto_login(add_additional_header)
            end_time = time.time()
            request_time = end_time - start_time
            if tee:
                print("Second request time =", request_time)

            # check if max requests exceeded
            request_iter = request_iter + 1
            if request_iter > max_request:
                break

            # extract
            if tee:
                print(request_result2)
            if bool(request_result2):
                if "status" in request_result2:
                    status = request_result2["status"]
                    if tee:
                        print(status)
                    if status == "PROCESSED" or status == "FAILED":
                        if "data" in request_result2:
                            return request_result2["data"]
                        else:
                            break
                    elif status == "IN QUEUE" or status == "IN PROGRESS":
                        if poll_time > request_time:
                            time.sleep(poll_time - request_time)
                        continue
                    else:
                        break
                else:
                    break
            else:
                break
        # TODO: raise error
        return dict()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from idaes.core.util.constants import Constants
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import StrEnum
from ..util import register_costing_parameter_block, make_capital_cost_var


class CrystallizerCostType(StrEnum):
    default = "default"
    mass_basis = "mass_basis"
    volume_basis = "volume_basis"


def build_crystallizer_cost_param_block(blk):

    blk.steam_pressure = pyo.Var(
        initialize=3,
        units=pyo.units.bar,
        doc="Steam pressure (gauge) for crystallizer heating: 3 bar default based on Dutta example",
    )

    blk.efficiency_pump = pyo.Var(
        initialize=0.7,
        units=pyo.units.dimensionless,
        doc="Crystallizer pump efficiency - assumed",
    )

    blk.pump_head_height = pyo.Var(
        initialize=1,
        units=pyo.units.m,
        doc="Crystallizer pump head height -  assumed, unvalidated",
    )

    # Crystallizer operating cost information from literature
    blk.fob_unit_cost = pyo.Var(
        initialize=675000,
        doc="Forced circulation crystallizer reference free-on-board cost (Woods, 2007)",
        units=pyo.units.USD_2007,
    )

    blk.ref_capacity = pyo.Var(
        initialize=1,
        doc="Forced circulation crystallizer reference crystal capacity (Woods, 2007)",
        units=pyo.units.kg / pyo.units.s,
    )

    blk.ref_exponent = pyo.Var(
        initialize=0.53,
        doc="Forced circulation crystallizer cost exponent factor (Woods, 2007)",
        units=pyo.units.dimensionless,
    )

    blk.iec_percent = pyo.Var(
        initialize=1.43,
        doc="Forced circulation crystallizer installed equipment cost (Diab and Gerogiorgis, 2017)",
        units=pyo.units.dimensionless,
    )

    blk.volume_cost = pyo.Var(
        initialize=16320,
        doc="Forced circulation crystallizer cost per volume (Yusuf et al., 2019)",
        units=pyo.units.USD_2007,  ## TODO: Needs confirmation, but data is from Perry apparently
    )

    blk.vol_basis_exponent = pyo.Var(
        initialize=0.47,
        doc="Forced circulation crystallizer volume-based cost exponent (Yusuf et al., 2019)",
        units=pyo.units.dimensionless,
    )

    blk.steam_cost = pyo.Var(
        initialize=0.004,
        units=pyo.units.USD_2018 / (pyo.units.meter**3),
        doc="Steam cost, Panagopoulos (2019)",
    )

    blk.NaCl_recovery_value = pyo.Var(
        initialize=0,
        units=pyo.units.USD_2018 / pyo.units.kg,
        doc="Unit recovery value of NaCl",
    )

    costing = blk.parent_block()
    costing.register_flow_type("steam", blk.steam_cost)
    costing.register_flow_type("NaCl", blk.NaCl_recovery_value)


def cost_crystallizer(blk, cost_type=CrystallizerCostType.default):
    """
    Function for costing the FC crystallizer by the mass flow of produced crystals.
    The operating cost model assumes that heat is supplied via condensation of saturated steam (see Dutta et al.)

    Args:
        cost_type: Option for crystallizer cost function type - volume or mass basis
    """
    if (
        cost_type == CrystallizerCostType.default
        or cost_type == CrystallizerCostType.mass_basis
    ):
        cost_crystallizer_by_crystal_mass(blk)
    elif cost_type == CrystallizerCostType.volume_basis:
        cost_crystallizer_by_volume(blk)
    else:
        raise ConfigurationError(
            f"{blk.unit_model.name} received invalid argument for cost_type:"
            f" {cost_type}. Argument must be a member of the CrystallizerCostType Enum."
        )


def _cost_crystallizer_flows(blk):
    blk.costing_package.cost_flow(
        pyo.units.convert(
            (
                blk.unit_model.magma_circulation_flow_vol
                * blk.unit_model.dens_mass_slurry
                * Constants.acceleration_gravity
                * blk.costing_package.crystallizer.pump_head_height
                / blk.costing_package.crystallizer.efficiency_pump
            ),
            to_units=pyo.units.kW,
        ),
        "electricity",
    )

    blk.costing_package.cost_flow(
        pyo.units.convert(
            (blk.unit_model.work_mechanical[0] / _compute_steam_properties(blk)),
            to_units=pyo.units.m**3 / pyo.units.s,
        ),
        "steam",
    )

    blk.costing_package.cost_flow(
        blk.unit_model.solids.flow_mass_phase_comp[0, "Sol", "NaCl"],
        "NaCl",
    )


@register_costing_parameter_block(
    build_rule=build_crystallizer_cost_param_block,
    parameter_block_name="crystallizer",
)
def cost_crystallizer_by_crystal_mass(blk):
    """
    Mass-based capital cost for FC crystallizer
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            (
                blk.costing_package.crystallizer.iec_percent
                * blk.costing_package.crystallizer.fob_unit_cost
                * (
                    sum(
                        blk.unit_model.solids.flow_mass_phase_comp[0, "Sol", j]
                        for j in blk.unit_model.config.property_package.solute_set
                    )
                    / blk.costing_package.crystallizer.ref_capacity
                )
                ** blk.costing_package.crystallizer.ref_exponent
            ),
            to_units=blk.costing_package.base_currency,
        )
    )
    _cost_crystallizer_flows(blk)


@register_costing_parameter_block(
    build_rule=build_crystallizer_cost_param_block,
    parameter_block_name="crystallizer",
)
def cost_crystallizer_by_volume(blk):
    """
    Volume-based capital cost for FC crystallizer
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            (
                blk.costing_package.crystallizer.volume_cost
                * (
                    (
                        pyo.units.convert(
                            blk.unit_model.volume_suspension
                            * (
                                blk.unit_model.height_crystallizer
                                / blk.unit_model.height_slurry
                            ),
                            to_units=(pyo.units.ft) ** 3,
                        )
                    )
                    / pyo.units.ft**3
                )
                ** blk.costing_package.crystallizer.vol_basis_exponent
            ),
            to_units=blk.costing_package.base_currency,
        )
    )
    _cost_crystallizer_flows(blk)


def _compute_steam_properties(blk):
    """
    Function for computing saturated steam properties for thermal heating estimation.

    Args:
        pressure_sat:   Steam gauge pressure in bar

    Out:
        Steam thermal capacity (latent heat of condensation * density) in kJ/m3
    """
    pressure_sat = blk.costing_package.crystallizer.steam_pressure
    # 1. Compute saturation temperature of steam: computed from El-Dessouky expression
    tsat_constants = [
        42.6776 * pyo.units.K,
        -3892.7 * pyo.units.K,
        1000 * pyo.units.kPa,
        -9.48654 * pyo.units.dimensionless,
    ]
    psat = (
        pyo.units.convert(pressure_sat, to_units=pyo.units.kPa)
        + 101.325 * pyo.units.kPa
    )
    temperature_sat = tsat_constants[0] + tsat_constants[1] / (
        pyo.log(psat / tsat_constants[2]) + tsat_constants[3]
    )

    # 2. Compute latent heat of condensation/vaporization: computed from Sharqawy expression
    t = temperature_sat - 273.15 * pyo.units.K
    enth_mass_units = pyo.units.J / pyo.units.kg
    t_inv_units = pyo.units.K**-1
    dh_constants = [
        2.501e6 * enth_mass_units,
        -2.369e3 * enth_mass_units * t_inv_units**1,
        2.678e-1 * enth_mass_units * t_inv_units**2,
        -8.103e-3 * enth_mass_units * t_inv_units**3,
        -2.079e-5 * enth_mass_units * t_inv_units**4,
    ]
    dh_vap = (
        dh_constants[0]
        + dh_constants[1] * t
        + dh_constants[2] * t**2
        + dh_constants[3] * t**3
        + dh_constants[4] * t**4
    )
    dh_vap = pyo.units.convert(dh_vap, to_units=pyo.units.kJ / pyo.units.kg)

    # 3. Compute specific volume: computed from Affandi expression (Eq 5)
    t_critical = 647.096 * pyo.units.K
    t_red = temperature_sat / t_critical  # Reduced temperature
    sp_vol_constants = [
        -7.75883 * pyo.units.dimensionless,
        3.23753 * pyo.units.dimensionless,
        2.05755 * pyo.units.dimensionless,
        -0.06052 * pyo.units.dimensionless,
        0.00529 * pyo.units.dimensionless,
    ]
    log_sp_vol = (
        sp_vol_constants[0]
        + sp_vol_constants[1] * (pyo.log(1 / t_red)) ** 0.4
        + sp_vol_constants[2] / (t_red**2)
        + sp_vol_constants[3] / (t_red**4)
        + sp_vol_constants[4] / (t_red**5)
    )
    sp_vol = pyo.exp(log_sp_vol) * pyo.units.m**3 / pyo.units.kg

    # 4. Return specific energy: density * latent heat
    return dh_vap / sp_vol



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a Ozone-AOP unit
operation.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.unit_models.zero_order.ozone_zo import OzoneZOData
from watertap.unit_models.zero_order.aop_addition_zo import AOPAdditionMixin

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("OzoneAOPZO")
class OzoneAOPZOData(OzoneZOData, AOPAdditionMixin):
    """
    Zero-Order model for a Ozone-AOP unit operation.
    """

    def build(self):
        super().build()

        self._tech_type = "ozone_aop"

        self.oxidant_dose = Var(
            self.flowsheet().time, units=pyunits.mg / pyunits.L, doc="Oxidant dosage"
        )

        self.chemical_flow_mass = Var(
            self.flowsheet().time,
            units=pyunits.kg / pyunits.s,
            bounds=(0, None),
            doc="Mass flow rate of oxidant solution",
        )

        self.ozone_toc_ratio = Var(
            self.flowsheet().time,
            units=pyunits.dimensionless,
            doc="Ratio of ozone to total organic carbon",
        )

        self.oxidant_ozone_ratio = Var(
            self.flowsheet().time,
            units=pyunits.dimensionless,
            doc="Ratio of oxidant to ozone",
        )

        self._fixed_perf_vars.append(self.oxidant_ozone_ratio)

        @self.Constraint(self.flowsheet().time, doc="Ozone/TOC ratio constraint")
        def ozone_toc_ratio_constraint(b, t):
            return b.ozone_toc_ratio[t] == 1 + pyunits.convert(
                b.concentration_time[t]
                / b.contact_time[t]
                / b.properties_in[t].conc_mass_comp["toc"],
                to_units=pyunits.dimensionless,
            )

        @self.Constraint(self.flowsheet().time, doc="Oxidant dose constraint")
        def oxidant_dose_constraint(b, t):
            return b.oxidant_dose[t] == pyunits.convert(
                b.oxidant_ozone_ratio[t]
                * b.ozone_toc_ratio[t]
                * b.properties_in[t].conc_mass_comp["toc"],
                to_units=pyunits.mg / pyunits.L,
            )

        @self.Constraint(self.flowsheet().time, doc="Oxidant mass flow constraint")
        def chemical_flow_mass_constraint(b, t):
            return b.chemical_flow_mass[t] == pyunits.convert(
                b.oxidant_dose[t] * b.properties_in[t].flow_vol,
                to_units=pyunits.kg / pyunits.s,
            )

        self._perf_var_dict["Oxidant Dosage (mg/L)"] = self.oxidant_dose
        self._perf_var_dict["Oxidant Flow (kg/s)"] = self.chemical_flow_mass
        self._perf_var_dict["Oxidant/Ozone Ratio"] = self.oxidant_ozone_ratio
        self._perf_var_dict["Ozone/TOC Ratio"] = self.ozone_toc_ratio

    @property
    def default_costing_method(self):
        return self.cost_ozonation_aop

    @staticmethod
    def cost_ozonation_aop(blk):
        """
        General method for costing ozonation with AOP. Capital cost is
        based on the inlet flowrate, dosage of ozone and flow rate of H2O2.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B, C, D, E, F = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "ozone_capital_a_parameter",
                "ozone_capital_b_parameter",
                "ozone_capital_c_parameter",
                "ozone_capital_d_parameter",
                "aop_capital_a_parameter",
                "aop_capital_b_parameter",
            ],
        )

        # Add cost variable
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        # Get costing term for ozone addition
        expr = blk.unit_model._get_ozone_capital_cost(blk, A, B, C, D)

        # Add costing term for AOP addition
        expr += blk.unit_model._get_aop_capital_cost(blk, E, F)

        blk.capital_cost_constraint = pyo.Constraint(expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.chemical_flow_mass[t0], "hydrogen_peroxide"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of an evaporation pond unit
model.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var, value
from idaes.core import declare_process_block_class
from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("EvaporationPondZO")
class EvaporationPondZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a evaporation pond unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        build_sido(self)
        constant_intensity(self)

        self._tech_type = "evaporation_pond"

        self.air_temperature = Var(
            self.flowsheet().time,
            initialize=298,
            units=pyunits.kelvin,
            doc="Air temperature",
        )

        self.solar_radiation = Var(
            self.flowsheet().time,
            units=pyunits.mJ / pyunits.m**2,
            doc="Daily solar radiation incident",
        )

        self.dike_height = Var(
            self.flowsheet().time, units=pyunits.ft, doc="Pond dike height"
        )

        self.evaporation_rate_adj_factor = Var(
            self.flowsheet().time,
            units=pyunits.dimensionless,
            doc="Factor to adjust evaporation rate of pure water",
        )

        self.evap_rate_calc_a_parameter = Var(
            self.flowsheet().time,
            units=pyunits.mm / pyunits.d,
            doc="Evaporation rate calculation parameter A",
        )

        self.evap_rate_calc_b_parameter = Var(
            self.flowsheet().time,
            units=pyunits.m**2 / pyunits.mJ,
            doc="Evaporation rate calculation parameter B",
        )

        self.evap_rate_calc_c_parameter = Var(
            self.flowsheet().time,
            units=pyunits.m**2 / pyunits.mJ,
            doc="Evaporation rate calculation parameter C",
        )

        self.adj_area_calc_a_parameter = Var(
            self.flowsheet().time,
            units=pyunits.acres,
            doc="Adjusted area calculation parameter A",
        )

        self.adj_area_calc_b_parameter = Var(
            self.flowsheet().time,
            units=pyunits.dimensionless,
            doc="Adjusted area calculation parameter B",
        )

        self._fixed_perf_vars.append(self.air_temperature)
        self._fixed_perf_vars.append(self.solar_radiation)
        self._fixed_perf_vars.append(self.dike_height)
        self._fixed_perf_vars.append(self.evaporation_rate_adj_factor)
        self._fixed_perf_vars.append(self.evap_rate_calc_a_parameter)
        self._fixed_perf_vars.append(self.evap_rate_calc_b_parameter)
        self._fixed_perf_vars.append(self.evap_rate_calc_c_parameter)
        self._fixed_perf_vars.append(self.adj_area_calc_a_parameter)
        self._fixed_perf_vars.append(self.adj_area_calc_b_parameter)

        self.area = Var(
            self.flowsheet().time,
            initialize=1,
            units=pyunits.acres,
            bounds=(0, None),
            doc="Pond area needed based on evaporation rate",
        )

        self.adj_area = Var(
            self.flowsheet().time, units=pyunits.acres, doc="Adjusted pond area needed"
        )

        self.evaporation_rate_pure = Var(
            self.flowsheet().time,
            units=pyunits.mm / pyunits.d,
            doc="Calculated evaporation rate of pure water",
        )

        self.evaporation_rate_salt = Var(
            self.flowsheet().time,
            units=(pyunits.gallons / pyunits.minute / pyunits.acre),
            doc="Pure water evaporation rate adjusted for salinity",
        )

        @self.Constraint(
            self.flowsheet().time, doc="Evaporation rate of pure water constraint"
        )
        def evap_rate_pure_constraint(b, t):
            air_temperature_C = pyunits.convert_temp_K_to_C(value(b.air_temperature[t]))
            return (
                b.evaporation_rate_pure[t]
                == b.evap_rate_calc_a_parameter[t]
                * (
                    b.evap_rate_calc_b_parameter[t] * air_temperature_C
                    + b.evap_rate_calc_c_parameter[t]
                )
                * b.solar_radiation[t]
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Adjusted evaporation rate for salinity constraint",
        )
        def evap_rate_salt_constraint(b, t):
            evap_rate_gal_min_acre = pyunits.convert(
                b.evaporation_rate_pure[t],
                to_units=(pyunits.gallons / pyunits.minute / pyunits.acre),
            )
            return (
                b.evaporation_rate_salt[t]
                == evap_rate_gal_min_acre * b.evaporation_rate_adj_factor[t]
            )

        @self.Constraint(self.flowsheet().time, doc="Base area constraint")
        def area_constraint(b, t):
            q_out = pyunits.convert(
                self.properties_byproduct[t].flow_vol,
                to_units=pyunits.gallon / pyunits.minute,
            )
            return q_out == b.evaporation_rate_salt[t] * b.area[t]

        @self.Constraint(self.flowsheet().time, doc="Adjusted area constraint")
        def area_adj_constraint(b, t):
            area = b.area[t] / pyunits.acres
            dike_ht = b.dike_height[t] / pyunits.ft
            return b.adj_area[t] == b.adj_area_calc_a_parameter[t] * area * (
                1 + (b.adj_area_calc_b_parameter[t] * dike_ht) / area**0.5
            )

        self._perf_var_dict["Evaporation rate (mm/d)"] = self.evaporation_rate_pure
        self._perf_var_dict["Pond area (acres)"] = self.adj_area
        self._perf_var_dict["Pond dike height (ft)"] = self.dike_height

    @property
    def default_costing_method(self):
        return self.cost_evaporation_pond

    @staticmethod
    def cost_evaporation_pond(blk):
        """
        General method for costing evaporation pond. Capital cost is based on the pond area and
        other pond construction parameters.
        """

        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        (
            A,
            B,
            C,
            D,
            E,
            liner_thickness,
            land_cost,
            land_clearing_cost,
        ) = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "cost_per_acre_a_parameter",
                "cost_per_acre_b_parameter",
                "cost_per_acre_c_parameter",
                "cost_per_acre_d_parameter",
                "cost_per_acre_e_parameter",
                "liner_thickness",
                "land_cost",
                "land_clearing_cost",
            ],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            blk.unit_model.adj_area[t0]
            * (
                A
                + B * liner_thickness
                + C * land_cost
                + D * land_clearing_cost
                + E * blk.unit_model.dike_height[t0]
            ),
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        blk.capital_cost_constraint = pyo.Constraint(expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.ui.fsapi import FlowsheetInterface
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1690.amo_1690 import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    add_costing,
)
from pyomo.environ import units as pyunits, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent


def export_to_ui():
    return FlowsheetInterface(
        name="AMO 1690",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):
    fs = flowsheet
    # --- Input data ---
    # Feed conditions
    exports.add(
        obj=fs.feed.flow_vol[0],
        name="Volumetric flow rate",
        ui_units=pyunits.gal / pyunits.day,
        display_units="gal/day",
        rounding=2,
        description="Inlet volumetric flow rate",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "tss"],
        name="TSS concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet total suspended solids concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "cod"],
        name="COD concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet COD concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "tkn"],
        name="TKN concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet TKN concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "acetic_acid"],
        name="Acetic acid concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet acetic acid concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "ammonium_as_nitrogen"],
        name="NH4 concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet ammonium concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )

    # Unit model data, cloth media filtration
    exports.add(
        obj=fs.cmf.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="Cloth media filtration",
        is_output=False,
    )
    exports.add(
        obj=fs.cmf.energy_electric_flow_vol_inlet,
        name="Electricity intensity",
        ui_units=pyunits.kWh / pyunits.m**3,
        display_units="kWh/m3",
        rounding=2,
        description="Specific electricity intensity relating the intensity to the volume of product",
        is_input=True,
        input_category="Cloth media filtration",
        is_output=False,
    )
    exports.add(
        obj=fs.cmf.removal_frac_mass_comp[0, "tss"],
        name="TSS removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Total suspended solids removal [g-TSS removed/g-TSS inlet]",
        is_input=True,
        input_category="Cloth media filtration",
        is_output=False,
    )
    exports.add(
        obj=fs.cmf.removal_frac_mass_comp[0, "cod"],
        name="COD removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="COD removal [g-COD removed/g-COD inlet]",
        is_input=True,
        input_category="Cloth media filtration",
        is_output=False,
    )
    exports.add(
        obj=fs.cmf.removal_frac_mass_comp[0, "tkn"],
        name="TKN removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="TKN removal [g-TKN removed/g-TKN inlet]",
        is_input=True,
        input_category="Cloth media filtration",
        is_output=False,
    )

    # Unit cost data, cloth media filtration
    exports.add(
        obj=fs.costing.cloth_media_filtration.sizing_cost[None],
        name="Cloth media filtration cost",
        ui_units=(fs.costing.base_currency * pyunits.day) / pyunits.gal,
        display_units="($ * day)/gal",
        rounding=0,
        description="Cloth media filtration capital cost parameter",
        is_input=True,
        input_category="Cloth media filtration costing",
        is_output=False,
    )

    # Unit model data, anaerobic digestion
    exports.add(
        obj=fs.ad.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="Anaerobic digestion",
        is_output=False,
    )
    exports.add(
        obj=fs.ad.reaction_conversion[0, "tss_reaction"],
        name="TSS conversion",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Total suspended solids conversion [g-TSS reacted/g-TSS inlet]",
        is_input=True,
        input_category="Anaerobic digestion",
        is_output=False,
    )
    exports.add(
        obj=fs.ad.reaction_conversion[0, "cod_reaction"],
        name="COD conversion",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="COD conversion [g-COD reacted/g-COD inlet]",
        is_input=True,
        input_category="Anaerobic digestion",
        is_output=False,
    )
    exports.add(
        obj=fs.ad.reaction_conversion[0, "tkn_reaction"],
        name="TKN conversion",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="TKN conversion [g-TKN reacted/g-TKN inlet]",
        is_input=True,
        input_category="Anaerobic digestion",
        is_output=False,
    )
    exports.add(
        obj=fs.ad.energy_electric_flow_vol_inlet,
        name="Electricity intensity",
        ui_units=pyunits.kWh / pyunits.m**3,
        display_units="kWh/m3",
        rounding=2,
        description="Specific electricity intensity relating the intensity to the volume of product",
        is_input=True,
        input_category="Anaerobic digestion",
        is_output=False,
    )
    exports.add(
        obj=fs.ad.biogas_tss_ratio,
        name="Biogas:TSS ratio",
        ui_units=pyunits.m**3 / pyunits.kg,
        display_units="m3/kg",
        rounding=2,
        description="Ratio of biogas volume to total suspended solids mass",
        is_input=True,
        input_category="Anaerobic digestion",
        is_output=False,
    )

    # Unit model data, membrane evaporator
    exports.add(
        obj=fs.me.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="Membrane evaporator",
        is_output=False,
    )
    exports.add(
        obj=fs.me.removal_frac_mass_comp[0, "ammonium_as_nitrogen"],
        name="NH4 removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Ammonium removal [g-NH4 removed/g-NH4 inlet]",
        is_input=True,
        input_category="Membrane evaporator",
        is_output=False,
    )
    exports.add(
        obj=fs.me.removal_frac_mass_comp[0, "acetic_acid"],
        name="Acetic acid removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Acetic acid removal [g-CH3COOH removed/g-CH3COOH inlet]",
        is_input=True,
        input_category="Membrane evaporator",
        is_output=False,
    )
    exports.add(
        obj=fs.me.energy_electric_flow_vol_inlet,
        name="Electricity intensity",
        ui_units=pyunits.kWh / pyunits.m**3,
        display_units="kWh/m3",
        rounding=2,
        description="Specific electricity intensity relating the intensity to the volume of product",
        is_input=True,
        input_category="Membrane evaporator",
        is_output=False,
    )
    exports.add(
        obj=fs.me.water_flux,
        name="Water flux",
        ui_units=pyunits.m / pyunits.hr,
        display_units="m/h",
        rounding=2,
        description="Water flux",
        is_input=True,
        input_category="Membrane evaporator",
        is_output=False,
    )

    # Unit cost data, membrane evaporator
    exports.add(
        obj=fs.costing.membrane_evaporator.membrane_cost[None],
        name="Membrane cost",
        ui_units=fs.costing.base_currency / pyunits.m**2,
        display_units="$/m2 of membrane",
        rounding=0,
        description="Membrane evaporator capital cost parameter",
        is_input=True,
        input_category="Membrane evaporator costing",
        is_output=False,
    )

    # System costing
    exports.add(
        obj=fs.costing.utilization_factor,
        name="Utilization factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Utilization factor - [annual use hours/total hours in year]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.TIC,
        name="Practical investment factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=1,
        description="Practical investment factor - [total investment cost/direct "
        "capital costs]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.plant_lifetime,
        name="Plant lifetime",
        ui_units=pyunits.year,
        display_units="years",
        rounding=1,
        description="Plant lifetime",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.wacc,
        name="Discount rate",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Discount rate used in calculating the capital annualization",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.maintenance_costs_percent_FCI,
        name="Fixed operating cost factor",
        ui_units=1 / pyunits.year,
        display_units="fraction/year",
        rounding=2,
        description="Fixed operating cost factor - [annual fixed operating cost/total "
        "investment cost]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.electricity_cost,
        name="Electricity cost",
        ui_units=fs.costing.base_currency / pyunits.kWh,
        display_units="$/kWh",
        rounding=3,
        description="Electricity cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.biogas_cost,
        name="Biogas cost",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3",
        rounding=3,
        description="Biogas cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.fertilizer_cost,
        name="Fertilizer cost",
        ui_units=fs.costing.base_currency / pyunits.kg,
        display_units="$/kg",
        rounding=3,
        description="Fertilizer cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )

    # Outlets
    exports.add(
        obj=fs.filtered_water.properties[0].flow_vol,
        name="Filtered water flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet filtered water flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.filtered_water.properties[0].conc_mass_comp["tss"],
        name="Filtered water TSS concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet filtered water total suspended solids concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.filtered_water.properties[0].conc_mass_comp["cod"],
        name="Filtered water COD concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet filtered water COD concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.filtered_water.properties[0].conc_mass_comp["tkn"],
        name="Filtered water TKN concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet filtered water TKN concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.ad_byproduct.properties[0].flow_vol,
        name="AD byproduct flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet anaerobic digestion byproduct flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.ad_byproduct.properties[0].conc_mass_comp["H2O"],
        name="AD byproduct water concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet AD byproduct water concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.ad_byproduct.properties[0].conc_mass_comp["tss"],
        name="AD byproduct TSS concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet AD byproduct total suspended solids concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.ad_byproduct.properties[0].conc_mass_comp["cod"],
        name="AD byproduct COD concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet AD byproduct COD concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.ad_byproduct.properties[0].conc_mass_comp["tkn"],
        name="AD byproduct TKN concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet AD byproduct TKN concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.me_byproduct.properties[0].flow_vol,
        name="Membrane evaporator byproduct flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet membrane evaporator byproduct flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.me_byproduct.properties[0].conc_mass_comp["acetic_acid"],
        name="Membrane evaporator byproduct acetic acid concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet membrane evaporator byproduct acetic acid concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.me_byproduct.properties[0].conc_mass_comp["ammonium_as_nitrogen"],
        name="Membrane evaporator byproduct ammonium concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet membrane evaporator byproduct ammonium concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.me_treated.properties[0].flow_vol,
        name="Membrane evaporator treated flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet membrane evaporator treated flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.me_treated.properties[0].conc_mass_comp["acetic_acid"],
        name="Membrane evaporator treated acetic_acid concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet membrane evaporator treated acetic acid concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.me_treated.properties[0].conc_mass_comp["ammonium_as_nitrogen"],
        name="Membrane evaporator treated ammonium concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=5,
        description="Outlet membrane evaporator treated ammonium concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )

    # System metrics
    exports.add(
        obj=fs.costing.LCOT,
        name="LCOT",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed water",
        rounding=3,
        description="Levelized cost of treatment including operating and capital costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.costing.LCOT_with_revenue,
        name="LCOT with revenue",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed water",
        rounding=3,
        description="Levelized cost of treatment including revenue of biogas and fertilizer",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.costing.LC_biogas,
        name="LCOB",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of biogas",
        rounding=3,
        description="Levelized cost of biogas including operating and capital costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.costing.LC_biogas_with_revenue,
        name="LCOB with revenue",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of biogas",
        rounding=3,
        description="Levelized cost of biogas including revenue of fertilizer",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.costing.LC_fertilizer,
        name="LCOF",
        ui_units=fs.costing.base_currency / pyunits.kg,
        display_units="$/kg of fertilizer",
        rounding=3,
        description="Levelized cost of fertilizer including operating and capital costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.costing.LC_fertilizer_with_revenue,
        name="LCOF with revenue",
        ui_units=fs.costing.base_currency / pyunits.kg,
        display_units="$/kg of fertilizer",
        rounding=3,
        description="Levelized cost of fertilizer including revenue of biogas",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )

    # Normalized metrics
    total_capital_norm = fs.costing.total_capital_cost / fs.feed.properties[0].flow_vol
    exports.add(
        obj=total_capital_norm,
        name="Total capital",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized total capital costs accounting for indirect "
        "capital and installation - [total capital costs/feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    direct_capital_norm = (
        fs.cmf.costing.direct_capital_cost
        + fs.ad.costing.direct_capital_cost
        + fs.me.costing.direct_capital_cost
    ) / fs.feed.properties[0].flow_vol
    exports.add(
        obj=direct_capital_norm,
        name="Direct capital",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized direct capital costs - [total direct capital "
        "costs/feed flow rate] ",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    elec_operating_norm = (
        fs.costing.aggregate_flow_costs["electricity"] / fs.costing.annual_water_inlet
    )
    exports.add(
        obj=elec_operating_norm,
        name="Electricity",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=2,
        description="Normalized electricity cost - [annual electricity costs/annual "
        "feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )

    # performance metrics
    recovery_vol = (
        fs.filtered_water.properties[0].flow_vol / fs.feed.properties[0].flow_vol
    )
    exports.add(
        obj=recovery_vol,
        name="Volumetric recovery",
        ui_units=pyunits.dimensionless,
        display_units="m3 of filtered water/m3 of feed",
        rounding=3,
        description="Normalized volumetric recovery",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_TSS = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["tss"]
        / fs.feed.properties[0].flow_mass_comp["tss"]
    )
    exports.add(
        obj=removal_TSS,
        name="TSS removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Total suspended solids removal fraction [1 - outlet TSS flow/inlet TSS flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_COD = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["cod"]
        / fs.feed.properties[0].flow_mass_comp["cod"]
    )
    exports.add(
        obj=removal_COD,
        name="COD removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="COD removal fraction [1 - outlet COD flow/inlet COD flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_TKN = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["tkn"]
        / fs.feed.properties[0].flow_mass_comp["tkn"]
    )
    exports.add(
        obj=removal_TKN,
        name="TKN removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="TKN removal fraction [1 - outlet TKN flow/inlet TKN flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    biogas_recovery_volume = fs.costing.utilization_factor * fs.ad.biogas_production[0]
    exports.add(
        obj=biogas_recovery_volume,
        name="Biogas recovery",
        ui_units=pyunits.m**3 / pyunits.year,
        display_units="m3-biogas/year",
        rounding=0,
        description="Biogas recovery volume",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    fertilizer_recovery_mass = fs.costing.utilization_factor * (
        fs.me_byproduct.flow_mass_comp[0, "ammonium_as_nitrogen"]
        + fs.me_byproduct.flow_mass_comp[0, "acetic_acid"]
    )
    exports.add(
        obj=fertilizer_recovery_mass,
        name="Fertilizer recovery",
        ui_units=pyunits.kg / pyunits.year,
        display_units="kg-fertilizer/year",
        rounding=0,
        description="Fertilizer recovery mass",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )

    # Capital costs
    exports.add(
        obj=fs.costing.total_capital_cost,
        name="Total",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Total capital costs - including investment factor to account "
        "for indirect capital",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.cmf.costing.capital_cost,
        name="Cloth media filtration",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Cloth media filtration",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.ad.costing.capital_cost,
        name="Anaerobic digestion",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Anaerobic digestion",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.me.costing.capital_cost,
        name="Membrane evaporator",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Membrane evaporator",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )

    # Operating costs
    exports.add(
        obj=fs.costing.total_operating_cost,
        name="Total",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Total annual operating costs - including electricity, heating, "
        "and fixed",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.aggregate_flow_costs["electricity"],
        name="Electricity",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual electricity costs ",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.total_fixed_operating_cost,
        name="Fixed",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual fixed operating costs - these costs include material "
        "replacement, maintenance, and labor",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )

    # Revenue
    total_revenue = (biogas_recovery_volume * fs.costing.biogas_cost) + (
        fertilizer_recovery_mass * fs.costing.fertilizer_cost
    )
    exports.add(
        obj=total_revenue,
        name="Total",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Total revenue - including the sale of struvite and purchase of MgCl2 and dry polymer",
        is_input=False,
        is_output=True,
        output_category="Revenue",
    )
    biogas_revenue = biogas_recovery_volume * fs.costing.biogas_cost
    exports.add(
        obj=biogas_revenue,
        name="Biogas",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Revenue from selling biogas",
        is_input=False,
        is_output=True,
        output_category="Revenue",
    )
    fertilizer_revenue = fertilizer_recovery_mass * fs.costing.fertilizer_cost
    exports.add(
        obj=fertilizer_revenue,
        name="Fertilizer",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Revenue from selling fertilizer",
        is_input=False,
        is_output=True,
        output_category="Revenue",
    )


def build_flowsheet(build_options=None, **kwargs):
    # build and solve initial flowsheet
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m)
    assert_optimal_termination(results)

    add_costing(m)
    assert_degrees_of_freedom(m, 0)
    m.fs.costing.initialize()

    results = solve(m)
    assert_optimal_termination(results)
    return m


def solve_flowsheet(flowsheet=None):
    fs = flowsheet
    results = solve(fs)
    return results



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from ..util import (
    register_costing_parameter_block,
    make_capital_cost_var,
)


def build_anaerobic_digestor_cost_param_block(blk):

    # NOTE: costing data are from NREL Waste-to-Energy Model
    blk.capital_a_parameter = pyo.Var(
        initialize=19.3552312e6,
        doc="A parameter for capital cost",
        units=pyo.units.USD_2012,
    )
    blk.capital_b_parameter = pyo.Var(
        initialize=0.6,
        doc="B parameter for capital cost",
        units=pyo.units.dimensionless,
    )
    blk.reference_flow = pyo.Var(
        initialize=911.054,
        doc="Reference flow for capital cost",
        units=pyo.units.m**3 / pyo.units.hr,
    )


@register_costing_parameter_block(
    build_rule=build_anaerobic_digestor_cost_param_block,
    parameter_block_name="anaerobic_digestor",
)
def cost_anaerobic_digestor(blk, cost_electricity_flow=True):
    """
    Anaerobic digestor costing method
    """
    cost_anaerobic_digestor_capital(
        blk,
        blk.costing_package.anaerobic_digestor.capital_a_parameter,
        blk.costing_package.anaerobic_digestor.capital_b_parameter,
        blk.costing_package.anaerobic_digestor.reference_flow,
    )

    t0 = blk.flowsheet().time.first()
    if cost_electricity_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.electricity_consumption[t0],
                to_units=pyo.units.kW,
            ),
            "electricity",
        )


def cost_anaerobic_digestor_capital(
    blk, capital_a_parameter, capital_b_parameter, reference_flow
):
    """
    Generic function for costing an anaerobic digestor system.
    """
    make_capital_cost_var(blk)

    blk.capital_a_parameter = pyo.Expression(expr=capital_a_parameter)
    blk.capital_b_parameter = pyo.Expression(expr=capital_b_parameter)
    blk.reference_flow = pyo.Expression(expr=reference_flow)

    flow_in = pyo.units.convert(
        blk.unit_model.liquid_phase.properties_in[0].flow_vol,
        to_units=pyo.units.m**3 / pyo.units.hr,
    )
    sizing_term = pyo.units.convert(
        flow_in / blk.reference_flow, to_units=pyo.units.dimensionless
    )

    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.capital_a_parameter * sizing_term**blk.capital_b_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Project setup with setuptools
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from pathlib import Path

cwd = Path(__file__).parent
long_description = (cwd / "README.md").read_text()

SPECIAL_DEPENDENCIES_FOR_RELEASE = [
    "idaes-pse>=2.3.0,<2.4.0rc0",  # from PyPI
]

SPECIAL_DEPENDENCIES_FOR_PRERELEASE = [
    # update with a tag from the nawi-hub/idaes-pse
    # when a version of IDAES newer than the latest stable release from PyPI
    # will become needed for the watertap development
    "idaes-pse>=2.3.0,<2.4.0rc0",
]

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name="watertap",
    url="https://github.com/watertap-org/watertap",
    version="0.11.0",
    description="WaterTAP modeling library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="NAWI team",
    license="BSD",
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="water systems, chemical engineering, process modeling, filtration, desalination, nawi",
    # just include watertap and everything under it
    packages=find_packages(
        include=("watertap*",),
    ),
    python_requires=">=3.8",
    install_requires=[
        # primary requirements for unit and property models
        # maintainers: switch to SPECIAL_DEPENDENCIES_FOR_RELEASE when cutting a release of watertap
        *SPECIAL_DEPENDENCIES_FOR_RELEASE,
        "pyomo>=6.6.1",  # (also needed for units in electrolyte database (edb))
        "pyyaml",  # watertap.core.wt_database
        # the following requirements are for the electrolyte database (edb)
        "pymongo>3",  # database interface
        "fastjsonschema",  # schema validation
        "click",  # command-line tools with Click
        # for parameter_sweep
        "h5py",
        "requests",
        "scipy",
        # for watertap.ui.api_model (though may be generally useful)
        "pydantic < 2",
        "numpy",
        "importlib-resources",
    ],
    extras_require={
        "testing": [
            "pytest",
            "json-schema-for-humans",
            "mongomock",
            "pandas",
            "nbmake",
            "nbconvert",
        ],
        "notebooks": [
            "jupyter",
            "ipykernel",
        ],
        "oli_api": [
            "requests",
            "cryptography",  # for encrypting OLI credentials
        ],
        "dev": [
            "nbsphinx",  # jupyter notebook support for sphinx
            "jinja2<3.1.0",  # see watertap-org/watertap#449
            "Sphinx==7.1.*",  # docs
            "sphinx_rtd_theme",  # docs
            "urllib3 < 2",  # see watertap-org/watertap#1021,
            # other requirements
            "linkify-it-py",
            "json-schema-for-humans",  # pretty JSON schema in HTML
            "black",  # code formatting
            # other requirements
            "pytest",  # test framework
            "pytest-cov",  # code coverage
            "mongomock",  # mongodb mocking for testing
            "nbmake",
        ],
    },
    package_data={  # Optional
        "": [
            "*.json",
            "*.yaml",
            "*.yml",
            "*.csv",
            "*.png",
        ],
        "watertap.tools.oli_api.tests": [
            "test.dbs",
        ],
    },
    entry_points={
        # add edb CLI commands
        "console_scripts": [
            "edb = watertap.edb.commands:command_base",
        ],
        "watertap.flowsheets": [
            "nf = watertap.examples.flowsheets.nf_dspmde.nf_ui",
            "nf_with_bypass = watertap.examples.flowsheets.nf_dspmde.nf_with_bypass_ui",
            "bsm2 = watertap.examples.flowsheets.case_studies.full_water_resource_recovery_facility.BSM2_ui",
            "metab = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.metab.metab_ui",
            "suboxic_ASM = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.suboxic_activated_sludge_process.suboxic_ASM_ui",
            "Magprex = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1575_magprex.magprex_ui",
            "biomembrane_filtration = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.biomembrane_filtration.biomembrane_filtration_ui",
            "ENR = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.electrochemical_nutrient_removal.electrochemical_nutrient_removal_ui",
            "CANDO_P = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1595_photothermal_membrane_candoP.amo_1595_ui",
            "supercritical_sludge_to_gas = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.supercritical_sludge_to_gas.supercritical_sludge_to_gas_ui",
            "PAA = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.peracetic_acid_disinfection.peracetic_acid_disinfection_ui",
            "AMO 1690 = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1690.amo_1690_ui",
            "HRCS = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1575_hrcs.hrcs_ui",
            "groundwater_treatment = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.groundwater_treatment.groundwater_treatment_ui",
            "dye_desalination = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.dye_desalination.dye_desalination_ui",
            "swine_wwt = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.swine_wwt.swine_wwt_ui",
            "GLSD anaerobic digestion = watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.GLSD_anaerobic_digester.GLSD_anaerobic_digestion_ui",
            "mvc = watertap.examples.flowsheets.mvc.mvc_single_stage_ui",
            "RO = watertap.examples.flowsheets.RO_with_energy_recovery.RO_with_energy_recovery_ui",
        ],
    },
)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of an iron and manganese removal unit.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("IronManganeseRemovalZO")
class IronManganeseRemovalZOData(ZeroOrderBaseData):
    """
    Zero-Order model for an iron and manganese removal unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "iron_and_manganese_removal"

        build_sido(self)

        self.air_water_ratio = Var(
            self.flowsheet().time,
            units=pyunits.dimensionless,
            doc="Ratio of air to water",
        )

        self.flow_basis = Var(
            self.flowsheet().time, units=pyunits.m**3 / pyunits.hour, doc="Flow basis"
        )

        self.air_flow_rate = Var(
            self.flowsheet().time,
            units=pyunits.m**3 / pyunits.hour,
            doc="Air flow rate",
        )

        self.electricity_intensity_parameter = Var(
            units=pyunits.hp / (pyunits.m**3 / pyunits.hour),
            doc="Constant in electricity intensity equation",
        )

        self.filter_surf_area = Var(
            units=pyunits.m**2, doc="Dual media filter surface area"
        )

        self.num_filter_units = Var(
            units=pyunits.dimensionless, doc="Number of dual media filter units"
        )

        self.electricity = Var(
            self.flowsheet().config.time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Power consumption of iron and manganese removal",
        )

        self.electricity_intensity = Var(
            self.flowsheet().config.time,
            units=pyunits.kWh / pyunits.m**3,
            doc="Specific energy consumption with respect to feed flowrate",
        )

        self._fixed_perf_vars.append(self.air_water_ratio)
        self._fixed_perf_vars.append(self.flow_basis)
        self._fixed_perf_vars.append(self.electricity_intensity_parameter)
        self._fixed_perf_vars.append(self.filter_surf_area)
        self._fixed_perf_vars.append(self.num_filter_units)

        @self.Constraint(self.flowsheet().config.time, doc="Air flow rate constraint")
        def air_flow_rate_constraint(b, t):
            return b.air_flow_rate[t] == b.air_water_ratio[t] * b.flow_basis[t]

        @self.Constraint(
            self.flowsheet().config.time, doc="Electricity intensity constraint"
        )
        def electricity_intensity_constraint(b, t):
            q_in = pyunits.convert(
                b.properties_in[t].flow_vol, to_units=pyunits.m**3 / pyunits.hour
            )
            return b.electricity_intensity[t] == pyunits.convert(
                b.electricity_intensity_parameter * b.air_flow_rate[t] / q_in,
                to_units=pyunits.kWh / pyunits.m**3,
            )

        @self.Constraint(
            self.flowsheet().config.time, doc="Power consumption constraint"
        )
        def electricity_constraint(b, t):
            q_in = pyunits.convert(
                b.properties_in[t].flow_vol, to_units=pyunits.m**3 / pyunits.hour
            )
            return b.electricity[t] == b.electricity_intensity[t] * q_in

        self._perf_var_dict["Power Consumption (kW)"] = self.electricity
        self._perf_var_dict[
            "Electricity intensity per Inlet Flowrate  (kWh/m3)"
        ] = self.electricity_intensity

    @property
    def default_costing_method(self):
        return self.cost_iron_and_manganese_removal

    @staticmethod
    def cost_iron_and_manganese_removal(blk, number_of_parallel_units=1):
        """
        General method for costing iron and manganese removal processes. Capital cost
        is based on the cost of air blower, backwash and dual media filter.
        This method also registers the electricity demand as a costed flow.
        Args:
            number_of_parallel_units (int, optional) - cost this unit as
                        number_of_parallel_units parallel units (default: 1)
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B, C, D, E, F = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "capital_blower_a_parameter",
                "capital_backwash_a_parameter",
                "capital_backwash_b_parameter",
                "capital_filter_a_parameter",
                "capital_filter_b_parameter",
                "flow_exponent",
            ],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        cost_blower = A

        cost_backwash = B + C * pyo.units.convert(
            blk.unit_model.filter_surf_area, to_units=pyo.units.ft**2
        )

        cost_filter = D + E * pyo.units.convert(
            blk.unit_model.filter_surf_area, to_units=pyo.units.ft**2
        )

        cost_total = pyo.units.convert(
            cost_blower + cost_backwash + cost_filter * blk.unit_model.num_filter_units,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        Q = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol,
            to_units=pyo.units.m**3 / pyo.units.hour,
        )

        sizing_term = Q / blk.unit_model.flow_basis[t0]

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        blk.unit_model._general_power_law_form(
            blk,
            cost_total,
            F,
            sizing_term,
            factor,
            number_of_parallel_units,
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
eNRTL property configuration dicts for synthetic hard water

References:

[1] Islam, R.I., et al., Molecular thermodynamics for scaling prediction: Case
of membrane distillation, Separation and Purification Technology, 2021,
Vol. 276.

Author: Andrew Lee
"""

from pyomo.environ import Param, units as pyunits

from idaes.core import AqueousPhase, Solvent, Apparent, Anion, Cation
from idaes.models.properties.modular_properties.eos.enrtl import ENRTL
from idaes.models.properties.modular_properties.eos.enrtl_reference_states import (
    Symmetric,
)
from idaes.models.properties.modular_properties.base.generic_property import StateIndex
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.pure.electrolyte import (
    relative_permittivity_constant,
)


class ConstantVolMol:
    def build_parameters(self):
        self.vol_mol_pure = Param(
            initialize=18e-6, units=pyunits.m**3 / pyunits.mol, mutable=True
        )

    def return_expression(self, cobj, T):
        return cobj.vol_mol_pure


configuration = {
    "components": {
        "H2O": {
            "type": Solvent,
            "vol_mol_liq_comp": ConstantVolMol,
            "relative_permittivity_liq_comp": relative_permittivity_constant,
            "parameter_data": {
                "mw": (18e-3, pyunits.kg / pyunits.mol),
                "relative_permittivity_liq_comp": 78.54,
            },
        },
        "NaCl": {"type": Apparent, "dissociation_species": {"Na_+": 1, "Cl_-": 1}},
        "Na2SO4": {"type": Apparent, "dissociation_species": {"Na_+": 2, "SO4_2-": 1}},
        "CaCl2": {"type": Apparent, "dissociation_species": {"Ca_2+": 1, "Cl_-": 2}},
        "CaSO4": {"type": Apparent, "dissociation_species": {"Ca_2+": 1, "SO4_2-": 1}},
        "MgCl2": {"type": Apparent, "dissociation_species": {"Mg_2+": 1, "Cl_-": 2}},
        "MgSO4": {"type": Apparent, "dissociation_species": {"Mg_2+": 1, "SO4_2-": 1}},
        "Na_+": {"type": Cation, "charge": +1},
        "Ca_2+": {"type": Cation, "charge": +2},
        "Mg_2+": {"type": Cation, "charge": +2},
        "Cl_-": {"type": Anion, "charge": -1},
        "SO4_2-": {"type": Anion, "charge": -2},
    },
    "phases": {
        "Liq": {
            "type": AqueousPhase,
            "equation_of_state": ENRTL,
            "equation_of_state_options": {"reference_state": Symmetric},
        }
    },
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    "state_definition": FTPx,
    "state_components": StateIndex.true,
    "pressure_ref": 1e5,
    "temperature_ref": 298.15,
    "parameter_data": {
        "Liq_tau": {  # Table 1 [1]
            ("H2O", "Na_+, Cl_-"): 8.866,
            ("Na_+, Cl_-", "H2O"): -4.451,
            ("H2O", "Ca_2+, Cl_-"): 10.478,
            ("Ca_2+, Cl_-", "H2O"): -5.231,
            ("H2O", "Mg_2+, Cl_-"): 10.854,
            ("Mg_2+, Cl_-", "H2O"): -5.409,
            ("H2O", "Na_+, SO4_2-"): 8.012,
            ("Na_+, SO4_2-", "H2O"): -3.903,
            ("H2O", "Ca_2+, SO4_2-"): 6.932,
            ("Ca_2+, SO4_2-", "H2O"): -3.466,
            ("H2O", "Mg_2+, SO4_2-"): 8.808,
            ("Mg_2+, SO4_2-", "H2O"): -4.383,
            ("Na_+, Cl_-", "Ca_2+, Cl_-"): -0.468,
            ("Ca_2+, Cl_-", "Na_+, Cl_-"): 0.41,
            ("Na_+, Cl_-", "Mg_2+, Cl_-"): -0.328,
            ("Mg_2+, Cl_-", "Na_+, Cl_-"): -0.981,
            ("Ca_2+, Cl_-", "Mg_2+, Cl_-"): 0.22,
            ("Mg_2+, Cl_-", "Ca_2+, Cl_-"): 0.322,
            ("Na_+, SO4_2-", "Ca_2+, SO4_2-"): -0.761,
            ("Ca_2+, SO4_2-", "Na_+, SO4_2-"): 0.368,
            ("Na_+, SO4_2-", "Mg_2+, SO4_2-"): -0.327,
            ("Mg_2+, SO4_2-", "Na_+, SO4_2-"): 0.799,
            ("Ca_2+, SO4_2-", "Mg_2+, SO4_2-"): 0,
            ("Mg_2+, SO4_2-", "Ca_2+, SO4_2-"): 0.383,
            ("Na_+, Cl_-", "Na_+, SO4_2-"): 0.803,
            ("Na_+, SO4_2-", "Na_+, Cl_-"): -0.634,
            ("Ca_2+, Cl_-", "Ca_2+, SO4_2-"): 0,
            ("Ca_2+, SO4_2-", "Ca_2+, Cl_-"): -0.264,
            ("Mg_2+, Cl_-", "Mg_2+, SO4_2-"): -0.707,
            ("Mg_2+, SO4_2-", "Mg_2+, Cl_-"): -0.841,
        }
    },
    "default_scaling_factors": {
        ("mole_frac_comp", "Na_+"): 1e2,
        ("mole_frac_comp", "Ca_2+"): 1e4,
        ("mole_frac_comp", "Mg_2+"): 1e3,
        ("mole_frac_comp", "SO4_2-"): 1e4,
        ("mole_frac_comp", "Cl_-"): 1e2,
        ("mole_frac_comp", "H2O"): 1,
        ("flow_mol", None): 1e-1,
    },
}
